/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   Position.h
 * Author: peter
 *
 * Created on 31. Jänner 2012, 18:23
 */

#ifndef POSITION_H
#define	POSITION_H

#include <string>
#include <iostream>
#include <stdio.h>
#include <cstdlib>
#include "main.h"

//! Class to Backup Move-Data
/*!
 * This class is used during alpha-beta-search to backup some
 * move-data (castling rules and en passant).
*/
struct backup {
    short int ep; /*!< stores possible en-passant of a move */
    char castlingRules; /*!< stores if move is a castling */
};

using namespace std;

//! Main Class of the Engine
/*!
 * This class handles chess-positions. Board situations are stored
 * with bitboards. The class also contains the move-generator and
 * make/undo-move-functions and various other basic functions.
 * All classes beginning with a name PositionXXX are based on this
 * class and represent different game-playing-modes.
*/
class Position {
public:
    bool onMove;
    int bestMoveIndex;
    unsigned int sum;
    unsigned long mvD[128][256][2];
    unsigned short int mvSize[128];
    
    Position();
    Position(const Position &orig);
    virtual ~Position() {};
    virtual string getName()=0;
    virtual void go(unsigned int depth)=0;
    void setFen(const string &fen);
    string getFen();
    int getMMChilds(string moveArray[], int moveIndex[]);
    int getMMChildsCatch(string moveArray[], int moveIndex[]);
    void printBoard();
    void getInfo();  
    string getMoveString(int i);
    unsigned short int move_figure (const unsigned long &from, const unsigned long &to, const int &promotion);
    void getDivide(int color, int curr_depth, const int &depth, const bool &modus, long &sum, long &total);
    void getChildsPseudo(int depth, long &sum);
    void getPerftInfo(int curr_depth, int depth, Perftres *perft);
    bool isKingWhiteLost();
    bool isKingBlackLost();
    bool isBlackToMove();
    bool isMate();
    bool isEndgame();
    bool hasQueen();
    bool hasBothRooks();
    bool hasBothBishops();
    bool hasBothKnights();
    unsigned long getAllUnsignedLong();
    unsigned long getKingUnsignedLong(bool onMove);
    unsigned long getPawnsUnsignedLong(bool onMove);
    unsigned long getQueenUnsignedLong(bool onMove);
    unsigned long getRooksUnsignedLong(bool onMove);
    unsigned long getBishopsUnsignedLong(bool onMove);
    unsigned long getKnightsUnsignedLong(bool onMove);
    char getCastlingRules();
    bool isPromotionPossible();
    bool isRochadePossible();
    bool isCapturePossible();
    
    static const unsigned long BIN[];    

protected:
    int cd;
    unsigned long aw;
    unsigned long ab;    
    unsigned long all;
    unsigned long kings[2];
    unsigned long pawns[2];
    unsigned long bishops[2];
    unsigned long knights[2];
    unsigned long queens[2];
    unsigned long rooks[2];  
    
    inline int bitScanForward(const unsigned long &bb) {
        if (bb)
                return __builtin_ffsll(bb)-1;
        else
            return 0;
//        const unsigned long debruijn64 = 0x07EDD5E59A4E28C2;
//        return Position::index64[((bb & -bb) * debruijn64) >> 58];
        
//        const int index64[64] = {
//            0, 47,  1, 56, 48, 27,  2, 60,
//           57, 49, 41, 37, 28, 16,  3, 61,
//           54, 58, 35, 52, 50, 42, 21, 44,
//           38, 32, 29, 23, 17, 11,  4, 62,
//           46, 55, 26, 59, 40, 36, 15, 53,
//           34, 51, 20, 43, 31, 22, 10, 45,
//           25, 39, 14, 33, 19, 30,  9, 24,
//           13, 18,  8, 12,  7,  6,  5, 63
//        };
//
//        
//        const unsigned long debruijn64 = 0x03f79d71b4cb0a89;
//           return index64[((bb ^ (bb-1)) * debruijn64) >> 58];        

    }
    inline int popCount(unsigned long x) { // too slow
//        int count = 0;
//        for (int i = 0; i < 64; i++, x >>= 1)
//                count += (int)x & 1;
//        return count;
        
      int c = 0;

      while (x) {
        c++;
        x &= x - 1;
      }
      return (c);        
        
    }
    void sortMoves(unsigned short int sortedIndizes[], int cd, int bestmoveHash);
    bool opponent_target_bitboard2(const unsigned long &bitboard);
    void calc_moves();
    void calcCaptures();
    void doBackup();
    void undoMove(const unsigned long &from, const unsigned long &to, const int &promotion, const unsigned short int &capture);

    
private:    
    short int ep;
    char castlingRules;
    backup myBackup[128];
      
    string get_hum(int k);
    inline void addPawnSingles_black(const int &z);
    inline void addPawnSingles_white(const int &z);
    inline void addPawnSingles_black_captures(const int &z);
    inline void addPawnSingles_white_captures(const int &z);    
    inline void rochade(const unsigned long &old_king_pos, const unsigned long &new_king_pos, const unsigned long &old_rook_pos, const unsigned long &new_rook_pos);
    inline void updateBitboardLongs();    
    inline void split_moves_direct(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &myFigures);
    inline void split_moves_direct_captures(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op);
    inline void split_moves__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op);
    inline void split_moves_captures__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op);
    
    static const unsigned short int castlingR1[];
    static const unsigned long promotionRow[];
    static const unsigned long TARGET_Nx[];
    static const unsigned int TARGET_Nx_FROM[];
    static const unsigned int TARGET_Nx_TO[];
    static const unsigned long TARGET_Ex[];
    static const unsigned int TARGET_Ex_FROM[];
    static const unsigned int TARGET_Ex_TO[];    
    static const unsigned long TARGET_Sx[];
    static const unsigned int TARGET_Sx_FROM[];
    static const unsigned int TARGET_Sx_TO[];        
    static const unsigned long TARGET_Wx[];
    static const unsigned int TARGET_Wx_FROM[];
    static const unsigned int TARGET_Wx_TO[];        
    static const unsigned long TARGET_NEx[];
    static const unsigned int TARGET_NEx_FROM[];
    static const unsigned int TARGET_NEx_TO[];       
    static const unsigned long TARGET_SEx[];
    static const unsigned int TARGET_SEx_FROM[];
    static const unsigned int TARGET_SEx_TO[];       
    static const unsigned long TARGET_SWx[];
    static const unsigned int TARGET_SWx_FROM[];
    static const unsigned int TARGET_SWx_TO[];       
    static const unsigned long TARGET_NWx[];
    static const unsigned int TARGET_NWx_FROM[];
    static const unsigned int TARGET_NWx_TO[];           
    static const unsigned long K_TARGET2x[];
    static const unsigned int K_TARGET2x_FROM[];
    static const unsigned int K_TARGET2x_TO[];
    static const unsigned long KN_TARGET3x[];
    static const unsigned int KN_TARGET3x_FROM[];
    static const unsigned int KN_TARGET3x_TO[];
    static const unsigned long TARGET_BLACK_V1[];
    static const unsigned long TARGET_BLACK_V2[];
    static const unsigned long TARGET_BLACK_V3[];
    static const unsigned long TARGET_BLACK_V4[];
    static const unsigned long TARGET_WHITE_V1[];
    static const unsigned long TARGET_WHITE_V2[];
    static const unsigned long TARGET_WHITE_V3[];
    static const unsigned long TARGET_WHITE_V4[];
    static const unsigned long PAWN_ATTACKS[2][64];    
};

#endif	/* POSITION_H */